/*
 * USE OF THE LIFEx SOFTWARE IS SUBJECT TO THE FOLLOWING TERMS AND CONDITIONS. 
 * IF YOU DO NOT ACCEPT THESE TERMS AND CONDITIONS, DO NOT DOWNLOAD THE SOFTWARE. 
 * NON-PROFESSIONAL USERS ARE NOTIFIED THAT THE SOFTWARE AND THE RISKS INHERENT TO ITS USE,
 * MODIFICATION AND/OR DEVELOPMENT AND REPRODUCTION ARE NOT COVERED BY ANY WARRANTY ON THE PART OF THE CEA. 
 * IT IS HIGHLY RECOMMENDED THAT YOU ONLY LOAD AND/OR USE THE SOFTWARE IF YOU HAVE SPECIALIZED IN-DEPTH INFORMATION TECHNOLOGY KNOW-HOW.
 * THE COMMISSARIAT A L’ENERGIE ATOMIQUE ET AUX ENERGIES ALTERNATIVES (FRENCH ATOMIC ENERGY AGENCY), 
 * a public scientific, technical and industrial establishment, 
 * having its registered office at 25 rue Leblanc – bâtiment le Ponant D – 75015 Paris cedex, 
 * hereinafter referred to as "CEA", hereby licenses the right to use the LIFEx software as per the terms and conditions set out below.  
 */
package bardp.protocol.protocolTexture;

import bardp.gui.FrameGui;
import bardp.img.controller.FrameController;
import bardp.img.coor.Coor;
import bardp.img.coor.CoorList;
import bardp.img.coor.CoorRange;
import bardp.protocol.protocolCalciumQuantitation.CalciumQuantitationSub;
import bardp.protocol.protocolTexture.calc.CalcIntensityDiscretizationSeries;
import bardp.protocol.protocolTexture.calc.CalcSpatialResamplingROI;
import bardp.protocol.protocolTexture.calc.CalcSpatialResamplingSeries;
import bardp.protocol.protocolTexture.calc.RoiKeepOneLargest;
import bardp.protocol.protocolTexture.traditional.RimController;
import bardp.protocol.protocolTexture.traditional.ConventionalTraditional;
import bardp.protocol.protocolTexture.traditional.ConvertionDoubleUnitYOfNativeSeries;
import bardp.protocol.protocolTexture.gui.TextureGlobalROIJPanel;
import bardp.protocol.protocolTexture.histo.HistogramController;
import bardp.protocol.protocolTexture.histo.HistogramAllSeriesGui;
import bardp.protocol.protocolTexture.params.TextureJPanelOutputParameters;
import bardp.protocol.protocolTexture.shape.CalcSphericity;
import bardp.protocol.protocolTexture.textureGLCM.TextureGLCMController;
import bardp.protocol.protocolTexture.textureGLRLM.TextureGLRLM;
import bardp.protocol.protocolTexture.textureGLZLM.TextureGLZLM;
import bardp.protocol.protocolTexture.textureNGLDM.TextureNGLDM;
import bardp.protocol.protocolTexture.traditional.ConvertionRawZYXUnitYOfNativeSeries;
import bardp.protocol.protocolTexture.traditional.DiscretizedTraditional;
import bardp.raw.RawSparseZYX;
import bardp.raw.RawTZYX;
import bardp.roi.controller.RoiController;
import bardp.roi.controller.RoiModel;
import bardp.session.patient.study.series.SeriesModel;
import bardp.session.patient.study.series.SeriesUnits;
import bardp.session.patient.study.series.load.loadDcm.PT_NM.PT_NM;
import bardp.utility.MessageGui;
import bardp.utility.ThreadExecution;
import bardp.utility.Utils;
import bardp.utility.WorkDir;
import java.util.ArrayList;
import java.util.logging.Level;
import javax.vecmath.Vector3d;

/**
 * <p>
 * Title: Local Image Feature Extraction (LIFEx)</p>
 * <p>
 * Description: </p>
 * <p>
 * Copyright: Copyright (c) 2016</p>
 * <p>
 * Company: Inserm/CNRS/CEA</p>
 * <p>
 * @author Christophe Nioche, Inserm at lifexsoft.org
 * @version 6.0
 */
public class TextureControllerSub {

    public TextureControllerSub(
            final RoiModel roiModel,
            final SeriesModel series,
            final int it,
            final TextureGlobalROIJPanel texJPanelOption,
            final HistogramAllSeriesGui histogramAllSeriesGUI,
            final MessageGui messageGUI
    ) {
        this.roiModel = roiModel;
        this.series = series;
        this.it = it;
        this.texJPanelOption = texJPanelOption;
        this.histogramAllSeriesGUI = histogramAllSeriesGUI;
        this.messageGUI = messageGUI;
    }

    /**
     * @return
     */
    public boolean run() {

        long start_time = System.currentTimeMillis();
        boolean isCheckTask = false;

        TextureJPanelOutputParameters texJPanelOutputParameters = texJPanelOption.texJFrameOutputParameters.texJPanelOutputParameters;

        try {

            //<editor-fold defaultstate="collapsed" desc="Intensity discretization parameters">
            this.intensityDiscretizationNbGreyLevels = texJPanelOption.getIntensityDiscretizationNbGreyLevels();
            this.intensityDiscretizationBinSize = texJPanelOption.getIntensityDiscretizationBinSize();
            //</editor-fold>

            /**
             * **************************************************************
             * ROI: 1) cropping + converting in Series coordinates
             * **************************************************************
             */
            //<editor-fold defaultstate="collapsed" desc="convert coorList (SeriesRef) -> (Series)">
            coorListCropOnSeries = RoiController.roiZYX.convertCoorListToSeries(
                    FrameController.seriesSelected,
                    roiModel.index
            );
            CoorRange boundsOfSeries = coorListCropOnSeries.getBounds();
            //<editor-fold defaultstate="collapsed" desc="log">
            FrameGui.LOGGER.log(
                    Level.INFO,
                    "Define CoorListCropOnSeries (raw):\n"
                    + "min: {0}\n"
                    + "max: {1}\n"
                    + "nb of voxels: {2}\n",
                    new Object[]{
                        boundsOfSeries.min.toString(),
                        boundsOfSeries.max.toString(),
                        coorListCropOnSeries.size()
                    }
            );
            //</editor-fold>

            // crop correcting of coorList with -this.boundsOfSeriesRef.min
            coorListCropOnSeries.forEach((coor) -> {
                coor.set(
                        coor.getZ() - boundsOfSeries.min.getZ(),
                        coor.getY() - boundsOfSeries.min.getY(),
                        coor.getX() - boundsOfSeries.min.getX()
                );
            });
            //</editor-fold>
            CoorRange boundsOfCoorListCropOnSeries = new CoorRange();
            boundsOfCoorListCropOnSeries.min.set(0, 0, 0);
            boundsOfCoorListCropOnSeries.max.set(
                    boundsOfSeries.max.getZ() - boundsOfSeries.min.getZ(),
                    boundsOfSeries.max.getY() - boundsOfSeries.min.getY(),
                    boundsOfSeries.max.getX() - boundsOfSeries.min.getX()
            );
            //<editor-fold defaultstate="collapsed" desc="log">
            FrameGui.LOGGER.log(
                    Level.INFO,
                    "Define CoorListCropOnSeries (crop):\n"
                    + "min: {0}\n"
                    + "max: {1}\n"
                    + "nb of voxels: {2}\n",
                    new Object[]{
                        boundsOfCoorListCropOnSeries.min.toString(),
                        boundsOfCoorListCropOnSeries.max.toString(),
                        coorListCropOnSeries.size()
                    }
            );
            //</editor-fold>

            // construct of mask (binary 1 || 0) on Series
            roiZYXSparseCropOfSeries = coorListCropOnSeries.createRawZYXSparse(series, (short) 1);
            //<editor-fold defaultstate="collapsed" desc="debugg + log">
            //2 debug lines :
            //System.out.println("QQQQ::"+roiZYXSparseCropOfSeries.toString());
            //for (Coor p: roiZYXSparseCropOfSeries.getCoorList()) {
            //    System.out.println("Ater2[" + p.toString() + "]=" + rawZYXCropOfSeries.getTZYX(p));
            //}
            /*
            SeriesModel.createSeriesForDebugFromRawZYXSparse("Dbg:RoiAvInterp", (short)1, roiZYXSparseCropOfSeries);
            FrameController.setActiveAll(series);
             */
            FrameGui.LOGGER.log(
                    Level.INFO,
                    "\nDefine roiZYXSparseCropOnSeriesRef:\n"
                    + "nb voxels: {0} voxels",
                    roiZYXSparseCropOfSeries.getSizeOfCoorList()
            );
            //</editor-fold> 

            /**
             * **************************************************************
             * Cropping Series and get one 'it' time ************************
             * **************************************************************
             */
            //<editor-fold defaultstate="collapsed" desc="cropping RawTZYX -> RawZYXFloatCrop">
            RawTZYX rawZYXCropNative = series.rawTZYX.copyAndCropZYX(
                    it,
                    boundsOfSeries.min,
                    boundsOfSeries.max
            );
            //
            // be carefull : it = 0 at this point
            //
            //</editor-fold>
            //<editor-fold defaultstate="collapsed" desc="log">
            FrameGui.LOGGER.log(
                    Level.INFO,
                    "\nDefine rawZYXCropOfSeries (crop):\n"
                    + " -> dim : nz{0} ny{1} nx{2} = tot of {3} voxels in volume\n",
                    new Object[]{
                        rawZYXCropNative.nz,
                        rawZYXCropNative.ny,
                        rawZYXCropNative.nx,
                        rawZYXCropNative.nz * rawZYXCropNative.ny * rawZYXCropNative.nx
                    }
            );
            //</editor-fold>
            //<editor-fold defaultstate="collapsed" desc="debugg">
            //for (Coor p : coorListCropOnSeries            ) {
            //    System.out.println("After1["+p.toString()+"]=" 
            //            + rawZYXCropOfSeries.getTZYX(p));
            //}
            //</editor-fold>

            //<editor-fold defaultstate="collapsed" desc="ZYX Spatial resampling paramaters">
            this.spatialResampling = new Vector3d(
                    texJPanelOption.getXSpatialResampling(),
                    texJPanelOption.getYSpatialResampling(),
                    texJPanelOption.getZSpatialResampling()
            );

            //</editor-fold>
            //<editor-fold defaultstate="collapsed" desc="roiDirectory + outputFilename">
            String roiDirectory = WorkDir.getROIValueDirectory(FrameController.seriesRef);
            bardp.utility.Utils.createDirectory(roiDirectory);

            String outputFilename = bardp.utility.Utils.prepDirectoryAndFile(roiDirectory,
                    Utils.prepFile(FrameController.seriesRef.fromStudy.fromPatient.patientName)
                    + "_" + Utils.prepFile(FrameController.seriesRef.fromStudy.fromPatient.patientID)
                    + "_Texture"
                    + "_" + Utils.prepFile(series.seriesDescription)
                    + "_" + Utils.prepFile(roiModel.title)
            );
            //</editor-fold>
            //</editor-fold>

            /**
             * **************************************************************
             * Spatial Resampling *******************************************
             * **************************************************************
             */
            // run Spatial Resampling of image
            rawZYXCropNative = (new CalcSpatialResamplingSeries()).run(
                    rawZYXCropNative,
                    spatialResampling
            );

            //<editor-fold defaultstate="collapsed" desc="debug">
            /* SeriesModel.createSeriesForDebugFromRawZYXFloat(
                "Dbg:Ser5Lag" + roiMain.roiModel.description,
                rawZYXCropOfSeries
            ); // for debug*/
            //</editor-fold>
            roiZYXSparseCropOfSeries = (new CalcSpatialResamplingROI()).run(
                    roiZYXSparseCropOfSeries,
                    spatialResampling
            );
            //<editor-fold defaultstate="collapsed" desc="log">
            FrameGui.LOGGER.log(
                    Level.INFO,
                    "nb of voxels on series CalcSpatialResamplingROI: {0} voxels",
                    roiZYXSparseCropOfSeries.getSizeOfCoorList()
            );
            //</editor-fold>

            // keep one cluster (more largest), modified roiZYXSparseCropOfSeries
            RoiKeepOneLargest roiKeepOneLargest = new RoiKeepOneLargest(
                    roiZYXSparseCropOfSeries,
                    roiModel
            );
            roiKeepOneLargest.run();
            int roiZYXSparseCropOfSeriesNbVoxels = roiZYXSparseCropOfSeries.getSizeOfCoorList();

            //<editor-fold defaultstate="collapsed" desc="log">
            FrameGui.LOGGER.log(
                    Level.INFO,
                    "nb of voxels on series RoiKeepOneLargest: {0} voxels",
                    roiZYXSparseCropOfSeriesNbVoxels
            );
            //</editor-fold>

            /**
             * **************************************************************
             * Units ********************************************************
             * **************************************************************
             */
            boolean isUnitSuvBw = series.seriesUnit.unitY == SeriesUnits.UNIT_SUV_BW;
            boolean isUnitSuvLbm = series.seriesUnit.unitY == SeriesUnits.UNIT_SUV_LBM;
            boolean isUnitSuvBsa = series.seriesUnit.unitY == SeriesUnits.UNIT_SUV_BSA;
            boolean isUnitSuvIbw = series.seriesUnit.unitY == SeriesUnits.UNIT_SUV_IBW;
            boolean isUnitkBqmL = series.seriesUnit.unitY == SeriesUnits.UNIT_kBqmL;

            //<editor-fold defaultstate="collapsed" desc="checking of parameters">
            this.intensityRescalingMinBound = Float.MAX_VALUE; // Real value DICOM (not SUV if TEP)
            this.intensityRescalingMaxBound = -Float.MAX_VALUE; // Real value DICOM (not SUV if TEP)

            // if relative resampling checked
            if (texJPanelOption.jRadioButtonRelativeMinMax.isSelected()) {
                //<editor-fold defaultstate="collapsed" desc="jRadioButtonRelativeMinMax.isSelected">
                float value;
                for (Coor coorSeriesRaw : roiZYXSparseCropOfSeries.getCoorList()) {
                    value = rawZYXCropNative.getTZYX(rawZYXCropNative.t, coorSeriesRaw);
                    if (value > intensityRescalingMaxBound) {
                        intensityRescalingMaxBound = value;
                    }
                    if (value < intensityRescalingMinBound) {
                        intensityRescalingMinBound = value;
                    }
                }

                if (isUnitSuvBw) {
                    intensityRescalingMinBound = PT_NM.convCpsVxToSuvBw(series, intensityRescalingMinBound);
                    intensityRescalingMaxBound = PT_NM.convCpsVxToSuvBw(series, intensityRescalingMaxBound);
                } else if (isUnitSuvLbm) {
                    intensityRescalingMinBound = PT_NM.convCpsVxToSuvLbm(series, intensityRescalingMinBound);
                    intensityRescalingMaxBound = PT_NM.convCpsVxToSuvLbm(series, intensityRescalingMaxBound);
                } else if (isUnitSuvBsa) {
                    intensityRescalingMinBound = PT_NM.convCpsVxToSuvBsa(series, intensityRescalingMinBound);
                    intensityRescalingMaxBound = PT_NM.convCpsVxToSuvBsa(series, intensityRescalingMaxBound);
                } else if (isUnitSuvIbw) {
                    intensityRescalingMinBound = PT_NM.convCpsVxToSuvIbw(series, intensityRescalingMinBound);
                    intensityRescalingMaxBound = PT_NM.convCpsVxToSuvIbw(series, intensityRescalingMaxBound);
                } else if (isUnitkBqmL) {
                    intensityRescalingMinBound = PT_NM.convCpsVxTokBqmL(series, intensityRescalingMinBound);
                    intensityRescalingMaxBound = PT_NM.convCpsVxTokBqmL(series, intensityRescalingMaxBound);
                }

                //</editor-fold>
            } else if (texJPanelOption.jRadioButtonRelativeSd.isSelected()) {
                //<editor-fold defaultstate="collapsed" desc="jRadioButtonRelativeSd.isSelected">
                float value;
                float mean = 0f;
                float delta;
                long n = 0;
                float m2 = 0f;
                float stdev;
                for (Coor coorSeriesRaw : roiZYXSparseCropOfSeries.getCoorList()) {
                    value = rawZYXCropNative.getTZYX(rawZYXCropNative.t, coorSeriesRaw);
                    n++;
                    delta = value - mean;
                    mean += delta / n;
                    m2 += delta * (value - mean);
                }
                stdev = (float) Math.sqrt(m2 / n);
                intensityRescalingMinBound = mean - 3 * stdev;
                intensityRescalingMaxBound = mean + 3 * stdev;

                if (isUnitSuvBw) {
                    intensityRescalingMinBound = PT_NM.convCpsVxToSuvBw(series, intensityRescalingMinBound);
                    intensityRescalingMaxBound = PT_NM.convCpsVxToSuvBw(series, intensityRescalingMaxBound);
                } else if (isUnitSuvLbm) {
                    intensityRescalingMinBound = PT_NM.convCpsVxToSuvLbm(series, intensityRescalingMinBound);
                    intensityRescalingMaxBound = PT_NM.convCpsVxToSuvLbm(series, intensityRescalingMaxBound);
                } else if (isUnitSuvBsa) {
                    intensityRescalingMinBound = PT_NM.convCpsVxToSuvBsa(series, intensityRescalingMinBound);
                    intensityRescalingMaxBound = PT_NM.convCpsVxToSuvBsa(series, intensityRescalingMaxBound);
                } else if (isUnitSuvIbw) {
                    intensityRescalingMinBound = PT_NM.convCpsVxToSuvIbw(series, intensityRescalingMinBound);
                    intensityRescalingMaxBound = PT_NM.convCpsVxToSuvIbw(series, intensityRescalingMaxBound);
                } else if (isUnitkBqmL) {
                    intensityRescalingMinBound = PT_NM.convCpsVxTokBqmL(series, intensityRescalingMinBound);
                    intensityRescalingMaxBound = PT_NM.convCpsVxTokBqmL(series, intensityRescalingMaxBound);
                }

                //</editor-fold>
            } else { // if absolute resampling checked
                //<editor-fold defaultstate="collapsed" desc="jRadioButtonAbsoluteMinMax.isSelected">
                intensityRescalingMinBound = texJPanelOption.getIntensityRescalingMinBound(); //be careful ! already in unitY: SUV if PET, SPECT...
                intensityRescalingMaxBound = texJPanelOption.getIntensityRescalingMaxBound(); //be careful ! already in unitY: SUV if PET, SPECT...
                //</editor-fold>
            }
            //</editor-fold>

            /**
             * **************************************************************
             * Intensity Discretization and Resampling **********************
             * **************************************************************
             */
            // define bounds of coorList of SeriesRaw
            this.boundsOfSeriesRef = this.roiZYXSparseCropOfSeries.getBounds((short) 1);
            //<editor-fold defaultstate="collapsed" desc="log">
            FrameGui.LOGGER.log(
                    Level.INFO,
                    "Define roiZYXSparseCropOfSeries (crop):\n"
                    + "min: {0}\n"
                    + "max: {1}\n"
                    + "nb of voxels: {2}\n",
                    new Object[]{
                        boundsOfSeriesRef.min.toString(),
                        boundsOfSeriesRef.max.toString(),
                        roiZYXSparseCropOfSeries.getSizeOfCoorList()
                    }
            );
            //</editor-fold>

            isCheckTask = isCheckTask();
            if (isCheckTask) {

                // setting seriesRaw with Unit (example: SUV)
                RawTZYX rawZYXCropNativeWithUnit = (new ConvertionRawZYXUnitYOfNativeSeries()).run(
                        series,
                        rawZYXCropNative,
                        roiZYXSparseCropOfSeries
                );
                
                double[] doubleZYXCropNativeWithUnit = (new ConvertionDoubleUnitYOfNativeSeries()).run(
                        series,
                        rawZYXCropNative,
                        roiZYXSparseCropOfSeries
                );
                
                RawTZYX rawZYXCropDiscreteWithUnit = (new CalcIntensityDiscretizationSeries()).run(
                        series, roiModel, this,
                        rawZYXCropNativeWithUnit, roiZYXSparseCropOfSeries,
                        intensityDiscretizationNbGreyLevels, intensityDiscretizationBinSize,
                        intensityRescalingMinBound, intensityRescalingMaxBound
                );
                rawZYXCropDiscretMaxValue = rawZYXCropDiscreteWithUnit.maxValue;
                rawZYXCropDiscretMinValue = rawZYXCropDiscreteWithUnit.minValue;

                //<editor-fold defaultstate="collapsed" desc="print debug">
                //System.out.println("before discretization");
                //rawZYXCropOfSeries.printZYXDebug();
                //System.out.println("after discretization");
                //rawZYXCropDiscrete.printZYXDebug();
                //</editor-fold>
                histogramAllSeriesGUI.addDataset(roiModel,
                        doubleZYXCropNativeWithUnit,
                        intensityRescalingMinBound,
                        intensityRescalingMaxBound
                );

                /**
                 * **************************************************************
                 * Running all **************************************************
                 * **************************************************************
                 */
                String dir = WorkDir.getROIValueDirectory(FrameController.seriesRef);
                Utils.createDirectory(dir);

                int nbCores = Runtime.getRuntime().availableProcessors();

                this.distanceWithNeighbours = texJPanelOutputParameters.jComboBoxDistance.getSelectedIndex() + 1; // 1 is for 0 selected roiIndex;

                calciumQuantitationSub = new CalciumQuantitationSub(
                        series, 
                        roiModel
                );
                
                conventionalTraditional = new ConventionalTraditional(
                        series, rawZYXCropNativeWithUnit, roiZYXSparseCropOfSeries,
                        messageGUI
                );

                discretizedTraditional = new DiscretizedTraditional(
                        series, rawZYXCropDiscreteWithUnit, roiZYXSparseCropOfSeries,
                        messageGUI
                );

                conventionalRim = new RimController(
                        rawZYXCropNativeWithUnit,
                        roiZYXSparseCropOfSeries,
                        (short) 1 // mandatory by construction
                );

                discretizedRim = new RimController(
                        rawZYXCropDiscreteWithUnit,
                        roiZYXSparseCropOfSeries,
                        (short) 1 // mandatory by construction
                );

                histogramController = new HistogramController(
                        roiModel,
                        series, doubleZYXCropNativeWithUnit,
                        intensityDiscretizationNbGreyLevels,
                        intensityRescalingMinBound, intensityRescalingMaxBound,
                        texJPanelOutputParameters.jCheckBoxHISTO_Show.isSelected(),
                        texJPanelOutputParameters.jCheckBoxHISTO_Save.isSelected(),
                        messageGUI
                );

                texSphericity = new CalcSphericity(
                        roiZYXSparseCropOfSeries,
                        messageGUI
                );

                texGLCM = new TextureGLCMController(
                        outputFilename,
                        texJPanelOutputParameters.jCheckBoxGLCM_Save.isSelected(),
                        distanceWithNeighbours,
                        rawZYXCropDiscreteWithUnit, roiZYXSparseCropOfSeries, boundsOfSeriesRef,
                        messageGUI
                );

                texGLRLM = new TextureGLRLM(
                        outputFilename,
                        texJPanelOutputParameters.jCheckBoxGLRLM_Save.isSelected(),
                        rawZYXCropDiscreteWithUnit, roiZYXSparseCropOfSeries, boundsOfSeriesRef,
                        messageGUI
                );

                texGLZLM = new TextureGLZLM(
                        outputFilename,
                        texJPanelOutputParameters.jCheckBoxGLZLM_Save.isSelected(),
                        rawZYXCropDiscreteWithUnit, roiZYXSparseCropOfSeries, boundsOfSeriesRef,
                        messageGUI, nbCores
                );

                texNGLDM = new TextureNGLDM(
                        outputFilename,
                        texJPanelOutputParameters.jCheckBoxNGLDM_Save.isSelected(),
                        rawZYXCropDiscreteWithUnit, roiZYXSparseCropOfSeries, boundsOfSeriesRef,
                        messageGUI
                );

                // memory info
                double freeMemory = Runtime.getRuntime().freeMemory() / 1024d / 1024d / 1024d; // o -> Go
                String mesg = "Free memory system is " + Utils.d2s(freeMemory, 1) + " GB.\n";
                FrameGui.LOGGER.log(Level.INFO, mesg);

                ArrayList<Runnable> tasks = new ArrayList<>();
                if (texJPanelOutputParameters.jCheckBoxConventional_CalciumAgatstonScore.isSelected()) {
                    tasks.add(calciumQuantitationSub);
                }
                tasks.add(conventionalTraditional);
                tasks.add(discretizedTraditional);
                tasks.add(histogramController);
                if (texJPanelOutputParameters.jCheckBoxConventional_RIMmin_mean_max.isSelected()) {
                    tasks.add(conventionalRim);
                    tasks.add(discretizedRim);
                }

                float roiVoxMin = roiZYXSparseCropOfSeries.nz > 1
                        ? TextureControllerSub.ROI_VOX_MIN_3D
                        : TextureControllerSub.ROI_VOX_MIN_2D;

                if (roiZYXSparseCropOfSeriesNbVoxels >= roiVoxMin) {
                    if (this.boundsOfSeriesRef.max.getZ() - this.boundsOfSeriesRef.min.getZ() >= 1) { // "if" may be not need
                        tasks.add(texSphericity);
                    }
                    tasks.add(texGLCM);
                    tasks.add(texGLRLM);
                    tasks.add(texNGLDM);
                    tasks.add(texGLZLM);

                } else {
                    //<editor-fold defaultstate="collapsed" desc="message too small">
                    texSphericity.setIsTooSmall();
                    texGLCM.setIsTooSmall();
                    texGLRLM.setIsTooSmall();
                    texNGLDM.setIsTooSmall();
                    texGLZLM.setIsTooSmall();
                    MessageGui messageGUIError = new MessageGui(
                            "Texture Protocol",
                            "The volume of " + roiModel.description
                            + " is too small (min " + roiVoxMin + " voxels) to calculated texture features.",
                            MessageGui.ERROR_MESSAGE
                    );
                    //</editor-fold>
                }

                ThreadExecution.runWithThreadsAndWait(tasks, 15);

                //</editor-fold> 
                //</editor-fold> 
                //<editor-fold defaultstate="collapsed" desc="degraded mode of texGLZLM">
                if (texGLZLM.outOfMemory) {
                    MessageGui messageGUIError = new MessageGui(
                            "Texture GLZLM",
                            "OutOfMemoryError: please add Xmx memory on launch",
                            MessageGui.INFORMATION_MESSAGE);

                    FrameGui.LOGGER.log(Level.INFO, "Try new processing of TexGLZLM with 1 processor...");
                    texGLZLM = new TextureGLZLM(outputFilename, texJPanelOutputParameters.jCheckBoxGLZLM_Save.isSelected(),
                            rawZYXCropDiscreteWithUnit, roiZYXSparseCropOfSeries, boundsOfSeriesRef, messageGUI, 1);
                    texGLZLM.run();

                    if (texGLZLM.outOfMemory) {
                        messageGUIError = new MessageGui(
                                "Texture GLZLM",
                                "OutOfMemoryError: please add Xmx memory on launch",
                                MessageGui.ERROR_MESSAGE);
                    }
                }
                //</editor-fold>

                //<editor-fold defaultstate="collapsed" desc="degraded mode of TexGLRLM">
                if (texGLRLM.outOfMemory) {
                    MessageGui messageGUIError = new MessageGui(
                            "Texture GLRLM",
                            "OutOfMemoryError: please add Xmx memory on launch",
                            MessageGui.INFORMATION_MESSAGE);

                    FrameGui.LOGGER.log(Level.INFO, "Try new processing of TexGLZLM with 1 processor...");
                    texGLRLM = new TextureGLRLM(outputFilename, texJPanelOutputParameters.jCheckBoxGLRLM_Save.isSelected(),
                            rawZYXCropDiscreteWithUnit, roiZYXSparseCropOfSeries, boundsOfSeriesRef, messageGUI);
                    texGLRLM.run();

                    if (texGLRLM.outOfMemory) {
                        messageGUIError = new MessageGui(
                                "Texture GLRLM",
                                "OutOfMemoryError: please add Xmx memory on launch",
                                MessageGui.ERROR_MESSAGE);
                    }
                }
                //</editor-fold>

            }
        } catch (Exception ex) {
            FrameGui.LOGGER.log(Level.INFO, "TexMainSub:: error {0} ", ex);
        }
        long end_time = System.currentTimeMillis();
        long difference = end_time - start_time;
        FrameGui.LOGGER.log(Level.INFO, "Time calculation of textural proceding: {0} sec", (difference / 1_000d));

        return isCheckTask;
    }

    @Override
    public String toString() {
        String text;

        text = "Series / ROI parameters:"
                + "\nSeries processing:\t" + series.seriesDescription
                + "\nSeries modality:\t\t" + series.modality
                + "\nVOI name:\t\t" + roiModel.title;

        text += "\n\nTextural parameters:"
                + "\nSpatial Resampling       -> spatial spacing:\t" + this.spatialResampling.toString()
                + "\nIntensity Rescaling      -> Min Bound:\t" + this.intensityRescalingMinBound
                + "\nIntensity Rescaling      -> Max Bound:\t" + this.intensityRescalingMaxBound
                + "\nIntensity Discretization -> Number of Grey Levels:\t" + intensityDiscretizationNbGreyLevels
                + "\nIntensity Discretization -> Bin Size:\t" + intensityDiscretizationBinSize;

        text += "\n\nTextural features:";
        text += "\n" + texGLCM.toString();
        text += "\n" + texGLRLM.toString();
        text += "\n" + texNGLDM.toString();
        text += "\n" + texGLZLM.toString();
        text += "\n" + texSphericity.toString();
        text += "\n" + histogramController.toString();
        text += "\n" + conventionalTraditional.toString();
        text += "\n" + conventionalRim.toString();

        return text;
    }

    /**
     * version 3.42 -  check input of texture processing
     * <p>
     * @return
     */
    private boolean isCheckTask() {

        //<editor-fold defaultstate="collapsed" desc="check intensityResampling min max values">
        if (FrameGui.getExecutionModel() != FrameGui.ExecutionModel.EXEC_SCRIPTING_TEXTURE) {
            if (texJPanelOption.jRadioButtonAbsoluteMinMax.isSelected()) {
                if (this.intensityRescalingMinBound == 0f
                        && this.intensityRescalingMaxBound == 0f) {
                    MessageGui messageGUI = new MessageGui(
                            "Texture Protocol",
                            "The bound (Min or Max) value is not defined.",
                            MessageGui.ERROR_MESSAGE);
                    return false;
                }

            }
        }
        //</editor-fold>

        //<editor-fold defaultstate="collapsed" desc="check intensityDiscretization NbGreyLevels value">
        if (FrameGui.getExecutionModel() != FrameGui.ExecutionModel.EXEC_SCRIPTING_TEXTURE) {
            /*if (this.intensityDiscretizationNbGreyLevels < 1) {
            MessageGUI messageGUI = new MessageGUI(
                    "Texture Protocol",
                    "The nbGreyLevels value is not defined or cannot be smaller than 1.",
                    MessageGUI.ERROR_MESSAGE);
            return false;
         }*/
            if (this.intensityDiscretizationNbGreyLevels == 0
                    && this.intensityDiscretizationBinSize == 0) {
                MessageGui messageGUI = new MessageGui(
                        "Texture Protocol",
                        "The nbGreyLevels value and the bin size are not defined. Define at least one of the both.",
                        MessageGui.ERROR_MESSAGE);
                return false;
            }
        }
        //</editor-fold>

        //<editor-fold defaultstate="collapsed" desc="check at least one ROI existing">
        if (roiModel == null) {
            MessageGui messageGUI = new MessageGui(
                    "Texture Protocol",
                    "No ROI found, please create at least one ROI and start again.",
                    MessageGui.ERROR_MESSAGE);
            return false;
        }
        //</editor-fold>

        return true;
    }

    public HistogramController histogramController;
    public CalciumQuantitationSub calciumQuantitationSub;
    public ConventionalTraditional conventionalTraditional;
    public RimController conventionalRim;
    public RimController discretizedRim;
    public DiscretizedTraditional discretizedTraditional;
    public CalcSphericity texSphericity;

    public TextureGLCMController texGLCM;
    public TextureGLRLM texGLRLM;
    public TextureNGLDM texNGLDM;
    public TextureGLZLM texGLZLM;

    public final RoiModel roiModel;
    private CoorList coorListCropOnSeries;
    private CoorRange boundsOfSeriesRef;

    public int distanceWithNeighbours;

    public double intensityDiscretizationNbGreyLevels;
    public double intensityDiscretizationBinSize;

    private double intensityRescalingMinBound;
    private double intensityRescalingMaxBound;

    public double rawZYXCropDiscretMaxValue;
    public double rawZYXCropDiscretMinValue;

    public Vector3d spatialResampling;

    /**
     * get information about clustering for csv user row
     */
    public String informationClusterStr = "";

    private final SeriesModel series;
    private final HistogramAllSeriesGui histogramAllSeriesGUI;
    private final int it;
    private final MessageGui messageGUI;
    private RawSparseZYX roiZYXSparseCropOfSeries;

    public final TextureGlobalROIJPanel texJPanelOption;

    /**
     * minimum size (vx) acceptable for 3D texture processing
     */
    static final public float ROI_VOX_MIN_3D = 64;

    /**
     * minimum size (vx) acceptable for 2D texture processing
     */
    static final public float ROI_VOX_MIN_2D = 16;

}
