/*
 * USE OF THE LIFEx SOFTWARE IS SUBJECT TO THE FOLLOWING TERMS AND CONDITIONS. 
 * IF YOU DO NOT ACCEPT THESE TERMS AND CONDITIONS, DO NOT DOWNLOAD THE SOFTWARE. 
 * NON-PROFESSIONAL USERS ARE NOTIFIED THAT THE SOFTWARE AND THE RISKS INHERENT TO ITS USE,
 * MODIFICATION AND/OR DEVELOPMENT AND REPRODUCTION ARE NOT COVERED BY ANY WARRANTY ON THE PART OF THE CEA. 
 * IT IS HIGHLY RECOMMENDED THAT YOU ONLY LOAD AND/OR USE THE SOFTWARE IF YOU HAVE SPECIALIZED IN-DEPTH INFORMATION TECHNOLOGY KNOW-HOW.
 * THE COMMISSARIAT A L’ENERGIE ATOMIQUE ET AUX ENERGIES ALTERNATIVES (FRENCH ATOMIC ENERGY AGENCY), 
 * a public scientific, technical and industrial establishment, 
 * having its registered office at 25 rue Leblanc – bâtiment le Ponant D – 75015 Paris cedex, 
 * hereinafter referred to as "CEA", hereby licenses the right to use the LIFEx software as per the terms and conditions set out below.  
 */
package bardp.protocol.protocolTexture.calc;

import bardp.gui.FrameGui;
import bardp.protocol.protocolTexture.textureGLCMMap.TextureGLCMMapController;
import bardp.raw.RawTZYX;
import bardp.session.patient.study.series.SeriesModel;
import bardp.session.patient.study.series.load.loadDcm.PT_NM.PT_NM;
import bardp.utility.MessageGui;
import bardp.utility.Utils;
import java.util.logging.Level;

/**
 * <p>
 * Title: Local Image Feature Extraction (LIFEx)</p>
 * <p>
 * Description: </p>
 * <p>
 * Copyright: Copyright (c) 2016</p>
 * <p>
 * Company: Inserm/CNRS/CEA</p>
 * <p>
 * @author Christophe Nioche, Inserm at lifexsoft.org
 * @version 6.0
 */
public class CalcMapIntensityDiscretizationSeries {

    /**
     * series discretized + if necessary -> convert in SUV (if PT, NM
     * modalities)
     * <br> with "intensityDiscretizationNbGreyLevels" ;
     * "intensityDiscretizationBinSize" if
     * intensityDiscretizationNbGreyLevels==0
     * <br> with "intensityDiscretizationBinSize" ;
     * "intensityDiscretizationNbGreyLevels" if intensityDiscretizationBinSize=0
     *
     * @param series
     * @param isPT_NM
     * @param texMapControllerSub
     * @param rawZYX
     * @param nbGreyLevels
     * @param binSize
     * @param boundMin
     * @param boundMax
     * @return
     */
    public RawTZYX run(
            final SeriesModel series,
            final boolean isPT_NM,
            final TextureGLCMMapController texMapControllerSub,
            final RawTZYX rawZYX,
            double nbGreyLevels,
            double binSize,
            double boundMin,
            double boundMax
    ) {
        double messageMinBound = -999_999_999d; // message display if value<boundMin
        double messageMaxBound = -999_999_999d; // message display if value>boundMax

        String msg = "\n";
        msg += "Intensity rescaling      -> bounds values (min <-> max): (" + boundMin + " <-> " + boundMax + ")\n";
        msg += "Intensity discretization -> nbGreyLevels: " + nbGreyLevels + "\n";
        msg += "Intensity discretization -> BinSize: " + binSize + "\n";

        //<editor-fold defaultstate="collapsed" desc="check parameters of intensity Discretization">
        if ((nbGreyLevels == 0) && (binSize == 0)) {
            // set default value
            nbGreyLevels = 64;
            msg += "both to zero -> nbGreyLevels applied by default: " + nbGreyLevels + "\n";
        }

        if (nbGreyLevels == 0) {
            nbGreyLevels = (boundMax - boundMin) / binSize;
            msg += "nbGreyLevels applied (corrected with (boundMax - boundMin) / binSize): " + nbGreyLevels + "\n";
        }
        texMapControllerSub.intensityDiscretizationNbGreyLevels = (float) nbGreyLevels; // new setting in this

        if (binSize == 0) {
            binSize = (boundMax - boundMin) / nbGreyLevels;
            msg += "binSize applied (corrected with (boundMax - boundMin) / nbGreyLevels): " + binSize + "\n";
        }
        texMapControllerSub.intensityDiscretizationBinSize = (float) binSize; // new setting in this
        //</editor-fold>
        
        float[][] dataZYX;
        float[][] discretZYX;
        RawTZYX rawZYXDiscret = rawZYX.createNewEmptyDataTZYX();

        for (int it = 0; it < rawZYX.nt; it++) {
            
            double value;
            double minValue = Double.MAX_VALUE;
            double maxValue = -Double.MAX_VALUE;
            double rangeMinMax = boundMax - boundMin;
            
            dataZYX = rawZYX.getZYX(it);
            discretZYX = rawZYXDiscret.getZYX(it);

            for (int iz = rawZYXDiscret.nz - 1; iz >= 0; iz--) {
                for (int iyx = rawZYXDiscret.nyx - 1; iyx >= 0; iyx--) {

                    // lecture des valeurs dans la serieRaw
                    value = dataZYX[iz][iyx];
                    if (isPT_NM) {
                        value = PT_NM.convCpsVxToSuvBw(series, value);
                    }
                    if (value < boundMin) {
                        messageMinBound = value;
                        value = boundMin;
                    }
                    if (value > boundMax) {
                        messageMaxBound = value;
                        value = boundMax;
                    }

                    value = (int) (nbGreyLevels * (value - boundMin) / rangeMinMax) + 1;

                    discretZYX[iz][iyx] = (float) value; // SUV unit if PET

                    if (value > maxValue) {
                        maxValue = value;
                    }
                    if (value < minValue) {
                        minValue = value;
                    }
                }
            }
            //</editor-fold>

            //<editor-fold defaultstate="collapsed" desc="error message on MinBound">
            if (messageMinBound != -999_999_999d) {
                MessageGui messageGUI = new MessageGui(
                        "Texture Protocol",
                        "The value (" + Utils.d2s(messageMinBound, 1) + ") of Map"
                        + " is lower than absolute min field value: " + Utils.d2s(boundMin, 1)
                        + " -> truncated value",
                        MessageGui.WARNING_MESSAGE);
            }
            //</editor-fold>

            //<editor-fold defaultstate="collapsed" desc="error message on MaxBound">
            if (messageMaxBound != -999_999_999d) {
                MessageGui messageGUI = new MessageGui(
                        "Texture Protocol",
                        "The value (" + Utils.d2s(messageMaxBound, 1) + ") of Map"
                        + " is higher than absolute max field value: " + Utils.d2s(boundMax, 1)
                        + " -> truncated value",
                        MessageGui.WARNING_MESSAGE);
            }
            //</editor-fold>

            if (maxValue == minValue) {
                MessageGui messageGUI = new MessageGui(
                        "Texture Protocol",
                        "[t" + it + "] "
                        + "After discretization, the min value (" + Utils.d2s(minValue, 2) + ") of Map"
                        + " is equal than max value: " + Utils.d2s(maxValue, 2)
                        + " -> texture canceled",
                        MessageGui.WARNING_MESSAGE);
            }

            // verification of discretization
            rawZYXDiscret.maxValueT[it] = (float)maxValue;
            rawZYXDiscret.minValueT[it] = (float)minValue;
            msg += "\n[t" + it + "] values after discretization (minValue=0 <-> maxValue=" + nbGreyLevels + "): "
                    + "(" + rawZYXDiscret.minValueT[it] + " <-> " + rawZYXDiscret.maxValueT[it] + ")";
            FrameGui.LOGGER.log(Level.INFO, msg);
        }
        return rawZYXDiscret;
    }
}
