/*
 * USE OF THE LIFEx SOFTWARE IS SUBJECT TO THE FOLLOWING TERMS AND CONDITIONS. 
 * IF YOU DO NOT ACCEPT THESE TERMS AND CONDITIONS, DO NOT DOWNLOAD THE SOFTWARE. 
 * NON-PROFESSIONAL USERS ARE NOTIFIED THAT THE SOFTWARE AND THE RISKS INHERENT TO ITS USE,
 * MODIFICATION AND/OR DEVELOPMENT AND REPRODUCTION ARE NOT COVERED BY ANY WARRANTY ON THE PART OF THE CEA. 
 * IT IS HIGHLY RECOMMENDED THAT YOU ONLY LOAD AND/OR USE THE SOFTWARE IF YOU HAVE SPECIALIZED IN-DEPTH INFORMATION TECHNOLOGY KNOW-HOW.
 * THE COMMISSARIAT A L’ENERGIE ATOMIQUE ET AUX ENERGIES ALTERNATIVES (FRENCH ATOMIC ENERGY AGENCY), 
 * a public scientific, technical and industrial establishment, 
 * having its registered office at 25 rue Leblanc – bâtiment le Ponant D – 75015 Paris cedex, 
 * hereinafter referred to as "CEA", hereby licenses the right to use the LIFEx software as per the terms and conditions set out below.  
 */
package bardp.protocol.protocolTexture.shape;

import bardp.session.patient.study.series.image.SpModel;
import bardp.raw.RawSparseZYX;

/**
 * <p>
 * Title: Local Image Feature Extraction (LIFEx)</p>
 * <p>
 * Description: </p>
 * <p>
 * Copyright: Copyright (c) 2016</p>
 * <p>
 * Company: Inserm/CNRS/CEA</p>
 * <p>
 * @author dlegland
 * @see inra.ijpb.measure.IntrinsicVolumes3D
 * @version 6.0
 */
public class InterfaceSurfaceAreaToDelete {

    public double run(
            final RawSparseZYX rawTZYX,
            final int roiIndex
    ) {
        double surface = 0d;

        //<editor-fold defaultstate="collapsed" desc="coordinates of voxels within tile">
        int[][] posList = new int[][]{
            {0, 0, 0},
            {1, 0, 0},
            {0, 1, 0},
            {1, 1, 0},
            {0, 0, 1},
            {1, 0, 1},
            {0, 1, 1},
            {1, 1, 1}
        };
        //</editor-fold>

        //<editor-fold defaultstate="collapsed" desc="mirrored within the tile">
        // flags indicating for each neighbor within tile and whether each
        // coordinate should be mirrored within the tile
        int[][] shifts = new int[][]{
            {1, 0, 0},
            {0, 1, 0},
            {0, 0, 1},
            {1, 1, 0},
            {1, 0, 1},
            {0, 1, 1},
            {1, 1, 1}
        };
        //</editor-fold>

        // pre-compute contributions
        double[] contribs = interceptsContributionsD13(rawTZYX.iopIppSp.spModel);
        int nNeighbors = shifts.length;
        int iNeighbors;

        // iterate over top-left-back corners of 2-by-2-by-2 configurations
        int iz, iy, ix, iPos;
        int x1, y1, z1, v1, x2, y2, z2, v2;
        int[] shift;
        int[] pos;
        for (iz = -1; iz <= rawTZYX.nz; iz++) { // -1 and <=nz for simulated bounds effect with getZYX(-1, -1, -1)=0
            for (iy = -1; iy <= rawTZYX.ny; iy++) {  // -1 and <=nz for simulated bounds effect with getZYX(-1, -1, -1)=0
                for (ix = -1; ix <= rawTZYX.nx; ix++) {  // -1 and <=nz for simulated bounds effect with getZYX(-1, -1, -1)=0

                    for (iPos = 0; iPos < posList.length; iPos++) { // compute absolute position of current voxel
                        pos = posList[iPos];
                        x1 = ix + pos[0];
                        y1 = iy + pos[1];
                        z1 = iz + pos[2];
                        v1 = rawTZYX.getZYX(z1, y1, x1); // value of tile
                        if (v1 != 0) {
                            continue;
                        }

                        // iterate over neighbors of current voxel within tile
                        for (iNeighbors = 0; iNeighbors < nNeighbors; iNeighbors++) {
                            // compute absolute position of neighbor
                            shift = shifts[iNeighbors];
                            x2 = shift[0] > 0 ? ix + (1 - pos[0]) : x1;
                            y2 = shift[1] > 0 ? iy + (1 - pos[1]) : y1;
                            z2 = shift[2] > 0 ? iz + (1 - pos[2]) : z1;
                            v2 = rawTZYX.getZYX(z2, y2, x2); // value of neighbor
                            if (v2 != roiIndex) {
                                continue;
                            }

                            // count a transition
                            surface += contribs[iNeighbors];
                        }
                    }
                }
            }
        }
        return surface;
    }

    /**
     * Return the contribution of each possible intercept.
     *
     * The resulting array has seven values, corresponding to the seven possible
     * neighbor voxel within a 2-by-2-by-2 configuration:
     * <ul>
     * <li>the intercept value in the X direction</li>
     * <li>the intercept value in the Y direction</li>
     * <li>the intercept value in the Z direction</li>
     * <li>the intercept value in the XY direction</li>
     * <li>the intercept value in the XZ direction</li>
     * <li>the intercept value in the YZ direction</li>
     * <li>the intercept value in the XYZ direction (cube diagonal)</li>
     * </ul>
     *
     * @param spModel the spatial calibration of the image
     * @return the contribution of each possible intercept.
     */
    private double[] interceptsContributionsD13(
            final SpModel spModel
    ) {
        // distances between a voxel and its neighbors.
        // di refer to orthogonal neighbors
        final double d1 = spModel.getXSpacing();
        final double d2 = spModel.getYSpacing();
        final double d3 = spModel.getZSpacing();
        final double vol = d1 * d2 * d3;

        // dij refer to neighbors on the same plane 
        final double d12 = Math.hypot(d1, d2);
        final double d13 = Math.hypot(d1, d3);
        final double d23 = Math.hypot(d2, d3);
        
        // dijk refer to the opposite voxel in a tile
        final double d123 = Math.hypot(d12, d3);

        /**
         * Number of directions for computing surface area or mean breadth with
         * Crofton Formula. Default is 13.
         */
        // direction weights corresponding to area of Voronoi partition on the
        // unit sphere, when germs are the 26 directions on the unit cube
        // Sum of (c1+c2+c3 + c4*2+c5*2+c6*2 + c7*4) equals 1.
        double[] weights = IntrinsicVolumes3DUtilsToDelete.directionWeights3d13(spModel);

        // contribution to number of intersections for each direction
        double[] kei = new double[7];
        kei[0] = 4 * weights[0] * vol / d1 / 8.0;
        kei[1] = 4 * weights[1] * vol / d2 / 8.0;
        kei[2] = 4 * weights[2] * vol / d3 / 8.0;
        kei[3] = 4 * weights[3] * vol / d12 / 4.0;
        kei[4] = 4 * weights[4] * vol / d13 / 4.0;
        kei[5] = 4 * weights[5] * vol / d23 / 4.0;
        kei[6] = 4 * weights[6] * vol / d123 / 2.0;

        return kei;
    }
}
