/*
 * USE OF THE LIFEx SOFTWARE IS SUBJECT TO THE FOLLOWING TERMS AND CONDITIONS. 
 * IF YOU DO NOT ACCEPT THESE TERMS AND CONDITIONS, DO NOT DOWNLOAD THE SOFTWARE. 
 * NON-PROFESSIONAL USERS ARE NOTIFIED THAT THE SOFTWARE AND THE RISKS INHERENT TO ITS USE,
 * MODIFICATION AND/OR DEVELOPMENT AND REPRODUCTION ARE NOT COVERED BY ANY WARRANTY ON THE PART OF THE CEA. 
 * IT IS HIGHLY RECOMMENDED THAT YOU ONLY LOAD AND/OR USE THE SOFTWARE IF YOU HAVE SPECIALIZED IN-DEPTH INFORMATION TECHNOLOGY KNOW-HOW.
 * THE COMMISSARIAT A L’ENERGIE ATOMIQUE ET AUX ENERGIES ALTERNATIVES (FRENCH ATOMIC ENERGY AGENCY), 
 * a public scientific, technical and industrial establishment, 
 * having its registered office at 25 rue Leblanc – bâtiment le Ponant D – 75015 Paris cedex, 
 * hereinafter referred to as "CEA", hereby licenses the right to use the LIFEx software as per the terms and conditions set out below.  
 */
package bardp.protocol.protocolTexture.textureGLCMMap;

import bardp.gui.FrameGui;
import bardp.raw.RawTZYX;
import bardp.session.patient.study.series.SeriesModel;
import bardp.utility.MessageGui;
import java.util.logging.Level;

/**
 * <p>
 * Title: Local Image Feature Extraction (LIFEx)</p>
 * <p>
 * Description: </p>
 * <p>
 * Copyright: Copyright (c) 2016</p>
 * <p>
 * Company: Inserm/CNRS/CEA</p>
 * <p>
 * @author Christophe Nioche, Inserm at lifexsoft.org
 * @version 6.0
 */
public class TextureGLCMMap implements Runnable {

    /**
     * processing the cooccurrence matrix map of grey level
     * <p>
     * @param rawZYXDiscret source image
     * @param nt
     * @param z
     * @param y
     * @param x
     * @param seriesModels
     * @param kernelSize
     * @param nbDirections
     * @param directions
     * @param messageGUI
     */
    public TextureGLCMMap(
            final RawTZYX rawZYXDiscret,
            final int nt,
            final int z,
            final int y,
            final int x,
            final SeriesModel[] seriesModels,
            final int kernelSize,
            final int nbDirections,
            final int[][] directions,
            final MessageGui messageGUI
    ) {
        this.rawZYXDiscret = rawZYXDiscret;
        this.nt = nt;
        this.z = z;
        this.y = y;
        this.x = x;
        this.seriesModels = seriesModels;
        this.kernelSize = kernelSize;
        this.nbDirections = nbDirections;
        this.directions = directions;
        this.messageGUI = messageGUI;
    }

    @Override
    public void run() {
        //<editor-fold defaultstate="collapsed" desc="variable definition">
        final double one = 1;
        final double epsilon = Math.ulp(1d);

        int greyValueRef, grayValueLocal;
        int sz, sy, sx;
        int iz, iy, iiy, ix;
        int nx;

        double homogeneity;
        double energy;
        double dissimilarity;
        double entropy_log10;
        //double correlation;
        double contrast;

        double diffrc;
        double value;
        double sum;

        int iDirection;
        int r, c;
        int nr, nc;
        double[][] matDirec;
        //</editor-fold>

        for (int it = 0; it < nt; it++) {

            // find maximum of grey level
            int maxValue = (int) Math.ceil(rawZYXDiscret.maxValueT[it]);
            int nbLevel = maxValue + 1;

            //<editor-fold defaultstate="collapsed" desc="variable definition">
            final double[][][] GLCM = new double[nbDirections][nbLevel][nbLevel];
            final double[][] trans_arr = new double[nbLevel][nbLevel];

            //</editor-fold>
            int distZ = (nbDirections == TextureGLCMMapController.NB_DIRECTIONS_2D)
                    ? 0 : (int) (kernelSize / 2f); // kernel3 => +-1, kernel5 => +-2, kernel7 => +-3, 
            int distYX = (int) (kernelSize / 2f); // kernel3 => +-1, kernel5 => +-2, kernel7 => +-3, 
            float[][] zyxDiscret;

            //<editor-fold defaultstate="collapsed" desc="GLCM core processing">
            zyxDiscret = rawZYXDiscret.getZYX(it);
            nx = rawZYXDiscret.nx;

            for (iz = z - distZ; iz <= z + distZ; iz++) {

                for (iy = y - distYX; iy <= y + distYX; iy++) {
                    iiy = iy * nx;

                    for (ix = x - distYX; ix <= x + distYX; ix++) {
                        try {

                            greyValueRef = (int) zyxDiscret[iz][iiy + ix];

                            for (iDirection = 0; iDirection < nbDirections; iDirection++) {

                                sz = iz + directions[2][iDirection]; // z;
                                sy = iy + directions[1][iDirection]; // y;
                                sx = ix + directions[0][iDirection]; // x;

                                grayValueLocal = (int) zyxDiscret[sz][sy * nx + sx];
                                GLCM[iDirection][greyValueRef][grayValueLocal]++;
                            }
                        } catch (Exception ex) {
                        }
                    }

                }
            }
            //</editor-fold>

            //<editor-fold defaultstate="collapsed" desc="Feature processing">
            homogeneity = 0;
            energy = 0;
            dissimilarity = 0;
            entropy_log10 = 0;
            // correlation = 0;
            contrast = 0;

            for (iDirection = 0; iDirection < nbDirections; iDirection++) {

                matDirec = GLCM[iDirection];
                nr = matDirec.length;
                nc = matDirec[0].length;

                //<editor-fold defaultstate="collapsed" desc="GLCM matrix transpose">
                for (r = 0; r < nbLevel; r++) {
                    for (c = 0; c < nbLevel; c++) {
                        trans_arr[r][c] = matDirec[c][r];
                    }
                }

                // sum of transpose
                for (r = 0; r < nbLevel; r++) {
                    for (c = 0; c < nbLevel; c++) {
                        matDirec[r][c] += trans_arr[r][c];
                    }
                }
                //</editor-fold>

                //<editor-fold defaultstate="collapsed" desc="GLCM normalisation processing">
                sum = 0;
                for (r = 0; r < nr; r++) {
                    for (c = 0; c < nc; c++) {
                        sum += matDirec[r][c];
                    }
                } // next step (/sum) in indice processing to optimized cpu
                //</editor-fold>

                //<editor-fold defaultstate="collapsed" desc="Indices processing">
                for (r = 0; r < nr; r++) {
                    for (c = 0; c < nc; c++) {

                        value = matDirec[r][c] / sum; // end of normalisation processing
                        diffrc = Math.abs(r - c);

                        homogeneity += value / (one + diffrc);
                        energy += value * value;
                        contrast += value * diffrc * diffrc;
                        dissimilarity += value * diffrc;
                        entropy_log10 += Math.abs(value * Math.log10(value + epsilon)); // against return -entropy
                    }
                }
                //  correlation += calculateCorrelation(GLCM[iDirection]); //TODO: more longer to calculate
                //</editor-fold>
            }

            homogeneity /= (double) nbDirections;
            energy /= (double) nbDirections;
            contrast /= (double) nbDirections;
            entropy_log10 /= (double) nbDirections;
            dissimilarity /= (double) nbDirections;
            //  correlation /= (double) nbDirections;
            
            //</editor-fold>

            seriesModels[0].rawTZYX.setTZYX(it, z, y, x, (float) homogeneity);
            seriesModels[1].rawTZYX.setTZYX(it, z, y, x, (float) energy);
            seriesModels[2].rawTZYX.setTZYX(it, z, y, x, (float) contrast);
            seriesModels[3].rawTZYX.setTZYX(it, z, y, x, (float) entropy_log10);
            seriesModels[4].rawTZYX.setTZYX(it, z, y, x, (float) dissimilarity);
            //seriesModels[5].rawTZYX.setTZYX(it, z, y, x, (float) correlation);
        }
        messageGUI.incProgress();
    }

    /**
     * correlation
     * <br>be careful, mat matrix must be normalized
     * <p>
     * @param mat
     * @return energy, Range = [-1 1] References: Haralick RM, Shapiro LG.
     * Computer and Robot Vision: Vol. 1, Addison-Wesley, 1992, p. 460. Bevk M,
     * Kononenko I. A Statistical Approach to Texture Description of Medical
     * Images: A Preliminary Study., The Nineteenth International Conference of
     * Machine Learning, Sydney, 2002.
     * http://www.cse.unsw.edu.au/~icml2002/workshops/MLCV02/MLCV02-Bevk.pdf,
     * p.3.
     */
    private double calculateCorrelation(
            final double[][] mat
    ) {
        // Correlation is defined as the covariance(r,c) / S(r)*S(c)
        // where S is the standard deviation.
        int r, c;
        int nr, nc;

        nr = mat.length;
        nc = mat[0].length;

        // Correlation is defined as the covariance(r,c) / S(r)*S(c)
        // where S is the standard deviation.
        //<editor-fold defaultstate="collapsed" desc="mr = Calculate the mean of a pixel value in the row">
        double mr = 0;
        for (c = 0; c < nc; c++) {
            for (r = 0; r < nr; r++) {
                mr += (r + 1) * mat[r][c];
            }
        }
        //</editor-fold>

        //<editor-fold defaultstate="collapsed" desc="sr = Calculate the standard deviation of a pixel value in the row">
        double sr = 0;
        for (c = 0; c < nc; c++) {
            for (r = 0; r < nr; r++) {
                sr += Math.pow((r + 1) - mr, 2) * mat[r][c];
            }
        }
        sr = Math.sqrt(sr);
        //</editor-fold>

        //<editor-fold defaultstate="collapsed" desc="mc = Calculate the mean of pixel value in the column">
        double mc = 0;
        for (c = 0; c < nc; c++) {
            for (r = 0; r < nr; r++) {
                mc += (c + 1) * mat[r][c];
            }
        }
        //</editor-fold>

        //<editor-fold defaultstate="collapsed" desc="sc = Calculate the standard deviation of pixel value in the column">
        double sc = 0;
        for (c = 0; c < nc; c++) {
            for (r = 0; r < nr; r++) {
                sc += Math.pow((c + 1) - mc, 2) * mat[r][c];
            }
        }
        sc = Math.sqrt(sc);
        //</editor-fold>

        //<editor-fold defaultstate="collapsed" desc="Calculate of covariance">
        double cov = 0;
        for (c = 0; c < nc; c++) {
            for (r = 0; r < nr; r++) {
                cov += ((r + 1) - mr) * ((c + 1) - mc) * mat[r][c];
            }
        }
        //</editor-fold>

        //<editor-fold defaultstate="collapsed" desc="Calculate of correlation">
        double cor = 0;
        try {
            cor = cov / (sr * sc);
        } catch (ArithmeticException ae) {
            FrameGui.LOGGER.log(Level.INFO, "GLCM_Correlation Processing / division by zero ... [ error ]");
        }
        //</editor-fold>
        return cor;
    }

    @Override
    public String toString() {
        return "\nFirst Order (GLCM Map)";
    }

    /**
     * discrete image
     */
    private final RawTZYX rawZYXDiscret;
    private final int nt;
    private final int z;
    private final int y;
    private final int x;

    private final SeriesModel[] seriesModels;

    /**
     * kernel size
     */
    private final int kernelSize;

    private final int nbDirections;
    private final int[][] directions;

    private final MessageGui messageGUI;
}
