/*
 * USE OF THE LIFEx SOFTWARE IS SUBJECT TO THE FOLLOWING TERMS AND CONDITIONS. 
 * IF YOU DO NOT ACCEPT THESE TERMS AND CONDITIONS, DO NOT DOWNLOAD THE SOFTWARE. 
 * NON-PROFESSIONAL USERS ARE NOTIFIED THAT THE SOFTWARE AND THE RISKS INHERENT TO ITS USE,
 * MODIFICATION AND/OR DEVELOPMENT AND REPRODUCTION ARE NOT COVERED BY ANY WARRANTY ON THE PART OF THE CEA. 
 * IT IS HIGHLY RECOMMENDED THAT YOU ONLY LOAD AND/OR USE THE SOFTWARE IF YOU HAVE SPECIALIZED IN-DEPTH INFORMATION TECHNOLOGY KNOW-HOW.
 * THE COMMISSARIAT A L’ENERGIE ATOMIQUE ET AUX ENERGIES ALTERNATIVES (FRENCH ATOMIC ENERGY AGENCY), 
 * a public scientific, technical and industrial establishment, 
 * having its registered office at 25 rue Leblanc – bâtiment le Ponant D – 75015 Paris cedex, 
 * hereinafter referred to as "CEA", hereby licenses the right to use the LIFEx software as per the terms and conditions set out below.  
 */
package bardp.protocol.protocolTexture.textureGLCMMap;

import bardp.gui.FrameGui;
import bardp.img.color.ColorController;
import bardp.img.color.ColorModelParam;
import bardp.img.coor.Coor;
import bardp.img.coor.CoorList;
import bardp.img.coor.CoorRange;
import bardp.img.controller.FrameController;
import bardp.img.picImage.PicImageBoxOptionController;
import bardp.protocol.protocolTexture.calc.CalcMapIntensityDiscretizationSeries;
import bardp.protocol.protocolTexture.gui.TextureLocalMapJPanel;
import bardp.raw.RawTZYX;
import bardp.roi.controller.RoiController;
import bardp.roi.controller.RoiMain;
import bardp.roi.curve.CurveController;
import bardp.roi.curve.CurveController.CurveEnum;
import bardp.session.patient.study.series.SeriesModel;
import bardp.session.patient.study.series.SeriesUnits;
import bardp.utility.MessageGui;
import bardp.utility.ThreadTask;
import bardp.utility.Utils;
import bardp.utility.WorkDir;
import java.util.concurrent.TimeUnit;
import java.util.logging.Level;

/**
 * <p>
 * Title: Local Image Feature Extraction (LIFEx)</p>
 * <p>
 * Description: </p>
 * <p>
 * Copyright: Copyright (c) 2016</p>
 * <p>
 * Company: Inserm/CNRS/CEA</p>
 * <p>
 * @author Christophe Nioche, Inserm at lifexsoft.org
 * @version 6.0
 */
public class TextureGLCMMapController extends ThreadTask {

    static public enum TexGLCMMapModel {

        /**
         * define execution on roi
         */
        ON_ROI,
        /**
         * define execution on bounding box on ROI
         */
        ON_BOUNDING_BOX_ON_ROI,
        /**
         * define execution on whole body
         */
        ON_WHOLE_BODY,
    };

    public TextureGLCMMapController(
            final SeriesModel series,
            final int kernelSize,
            final TexGLCMMapModel model,
            final TextureLocalMapJPanel texGLCMMapJPanelOption
    ) {
        this.series = series;
        this.kernelSize = kernelSize;
        this.model = model;
        this.texGLCMMapJPanelOption = texGLCMMapJPanelOption;
    }

    @Override
    public void taskCanceled() {
        taskEnd();
        MessageGui messagGui = new MessageGui(
                "Canceled operation",
                "Not supported yet.",
                MessageGui.WARNING_MESSAGE
        );
    }

    @Override
    public void taskEnd() {
        this.texGLCMMapJPanelOption.jButtonRun.setEnabled(true);
    }

    /**
     * add SeriesModel into Series GUI for debug only
     */
    @Override
    public void run() {

        this.texGLCMMapJPanelOption.jButtonRun.setEnabled(false);

        //<editor-fold defaultstate="collapsed" desc="chek selected ROI if necessary">
        RoiMain roiMain = null;
        if (model == TexGLCMMapModel.ON_ROI
                || model == TexGLCMMapModel.ON_BOUNDING_BOX_ON_ROI) {
            roiMain = RoiController.roiMainActived;
            if (roiMain == null) {
                new MessageGui("No ROI was selected !!",
                        "Select one ROI and try again.",
                        MessageGui.WARNING_MESSAGE
                );
                taskEnd();
                return;
            }
        }
        //</editor-fold>

        long start_time = System.currentTimeMillis();

        try {
            FrameController.setActivatedSeriesOnAllFrames(FrameController.seriesSelected);

            //<editor-fold defaultstate="collapsed" desc="Intensity discretization parameters">
            this.intensityDiscretizationNbGreyLevels = texGLCMMapJPanelOption.getIntensityDiscretizationNbGreyLevels();
            this.intensityDiscretizationBinSize = texGLCMMapJPanelOption.getIntensityDiscretizationBinSize();
            //</editor-fold>

            this.isPT_NM = series.modality.equals("PT")
                    || series.modality.equals("NM");

            //<editor-fold defaultstate="collapsed" desc="Intensity Rescaling Min/Max bound parameters">
            //<editor-fold defaultstate="collapsed" desc="jRadioButtonAbsoluteMinMax.isSelected">
            intensityRescalingMinBound = texGLCMMapJPanelOption.getIntensityRescalingMinBound(); // attention ! deja en SUV si PET, SPECT...
            intensityRescalingMaxBound = texGLCMMapJPanelOption.getIntensityRescalingMaxBound(); // attention ! deja en SUV si PET, SPECT...
            //</editor-fold>
            //</editor-fold>

            /**
             * **************************************************************
             * Intensity Discretization and Resampling
             * **************************************************************
             */
            if (isCheckTask()) {

                // setting continue serieRaw
                RawTZYX rawZYXCropDiscret = (new CalcMapIntensityDiscretizationSeries()).run(
                        series,
                        isPT_NM,
                        this,
                        series.rawTZYX,
                        intensityDiscretizationNbGreyLevels, intensityDiscretizationBinSize,
                        intensityRescalingMinBound, intensityRescalingMaxBound
                );

                /**
                 * **************************************************************
                 * Running all
                 * **************************************************************
                 */
                String dir = WorkDir.getROIValueDirectory(FrameController.seriesRef);
                Utils.createDirectory(dir);

                String[] seriesNames = new String[]{
                    "GLCM_Homogeneity",
                    "GLCM_Energy",
                    "GLCM_Contrast",
                    "GLCM_Entropy_log10",
                    "GLCM_Dissimilarity"
                //no   "GLCM_Correlation",
                };

                SeriesModel[] seriesModels = new SeriesModel[seriesNames.length];

                for (int i = 0; i < seriesNames.length; i++) {
                    //<editor-fold defaultstate="collapsed" desc="create seriesMap">
                    seriesModels[i] = series.createNewSeriesCalc(
                            "OT",
                            "k" + kernelSize + seriesNames[i],
                            series.rawTZYX.nt
                    );
                    boolean isVisible = seriesNames[i].equals("GLCM_Entropy_log10");
                    PicImageBoxOptionController.getPicImageBoxOptionAllPlanFromSerie(seriesModels[i]).picImageJPanelMenu.setVisibleEye(isVisible);
                    seriesModels[i].seriesUnit.unitY = SeriesUnits.UNIT_default3d;
                    PicImageBoxOptionController.updateAllMenuPicImageBoxOptionFromSerie(seriesModels[i]);
                    //</editor-fold>
                }

                //<editor-fold defaultstate="collapsed" desc="nbDirections">
                int nbDirections;
                int[][] directions;
                if (rawZYXCropDiscret.nz >= 2) {
                    nbDirections = NB_DIRECTIONS_3D;
                    directions = DIRECTIONS_3D;
                } else {
                    nbDirections = NB_DIRECTIONS_2D;
                    directions = DIRECTIONS_2D;
                }
                //</editor-fold>

                //<editor-fold defaultstate="collapsed" desc="coorList of MAP processing voxels">
                CoorList coorList = new CoorList();
                int iz, iy, ix;

                if (null != model) {
                    switch (model) {
                        case ON_BOUNDING_BOX_ON_ROI:
                            // on reference series
                            //CoorRange rangeBounds = RoiController.roiZYX.getBounds(roiMain.roiModel.index);
                            //<editor-fold defaultstate="collapsed" desc="convert coorList (SeriesRef) -> (Series)">
                            coorList = RoiController.roiZYX.convertCoorListToSeries(
                                    FrameController.seriesSelected,
                                    roiMain.roiModel.index
                            );
                            //</editor-fold>

                            CoorRange boundsOfSeries = coorList.getBounds();
                            boundsOfSeries.max.set(
                                    boundsOfSeries.max.getZ()+1,
                                    boundsOfSeries.max.getY()+1,
                                    boundsOfSeries.max.getX()+1
                                    );
                            
                            for (iz = boundsOfSeries.min.getZ(); iz <= boundsOfSeries.max.getZ(); iz++) {
                                for (iy = boundsOfSeries.min.getY(); iy <= boundsOfSeries.max.getY(); iy++) {
                                    for (ix = boundsOfSeries.min.getX(); ix <= boundsOfSeries.max.getX(); ix++) {
                                        coorList.add(new Coor(iz, iy, ix));
                                    }
                                }
                            }
                            break;

                        case ON_ROI:
                            // on reference series
                            //coorList = RoiController.roiZYX.getCoorList(roiMain.roiModel.index);

                            //<editor-fold defaultstate="collapsed" desc="convert coorList (SeriesRef) -> (Series)">
                            coorList = RoiController.roiZYX.convertCoorListToSeries(
                                    FrameController.seriesSelected,
                                    roiMain.roiModel.index
                            );
                            //</editor-fold>
                            break;

                        case ON_WHOLE_BODY:
                            // not execute on whole body, but only on value > 0.05
                            double thresholProcessing = 0.05;
                            for (iz = 0; iz <= series.rawTZYX.nz; iz++) {
                                for (iy = 0; iy <= series.rawTZYX.ny; iy++) {
                                    for (ix = 0; ix <= series.rawTZYX.nx; ix++) {
                                        if (series.rawTZYX.getTZYX(series.rawTZYX.t, iz, iy, ix) > thresholProcessing) {
                                            coorList.add(new Coor(iz, iy, ix));
                                        }
                                    }
                                }
                            }
                            break;

                        default:
                            break;
                    }
                }
                //</editor-fold>

                
                
                //<editor-fold defaultstate="collapsed" desc="log">
                FrameGui.LOGGER.log(
                        Level.INFO,
                        "Define CoorListCropOnSeries (raw):\n"
                        + "nb of voxels: {2}\n",
                        new Object[]{
                            coorList.size()
                        }
                );
                //</editor-fold>

                //<editor-fold defaultstate="collapsed" desc="MessageGUI: Textural processing on-going">
                messageGui = new MessageGui(
                        this,
                        "Textural Map",
                        "Textural Map processing on-going",
                        MessageGui.WAIT_MESSAGE,
                        100f / (float) coorList.size()
                );
                //</editor-fold>

                for (Coor coor : coorList) {
                    this.submit(
                            new TextureGLCMMap(
                                    rawZYXCropDiscret,
                                    series.rawTZYX.nt,
                                    coor.getZ(), coor.getY(), coor.getX(),
                                    seriesModels,
                                    kernelSize,
                                    nbDirections,
                                    directions,
                                    messageGui
                            )
                    );
                }

                this.shutdownAndAwaitTermination(5, TimeUnit.HOURS);

                // memory info
                double freeMemory = Runtime.getRuntime().freeMemory() / 1024d / 1024d / 1024d; // o -> Go
                String mesg = "Free memory system is " + Utils.d2s(freeMemory, 1) + " GB.\n";
                FrameGui.LOGGER.log(Level.INFO, mesg);

                //<editor-fold defaultstate="collapsed" desc="message break">
                if (isBreak) {
                    for (int i = 0; i < seriesNames.length; i++) {
                        FrameGui.session.removePatient(seriesModels[i].fromStudy.fromPatient);
                    }
                    messageGui.end();

                    MessageGui m = new MessageGui(
                            "Texture Map",
                            "Processing aborted by user",
                            MessageGui.WARNING_MESSAGE);
                    taskEnd();
                    return;
                }
                //</editor-fold>

                //<editor-fold defaultstate="collapsed" desc="message OK">
                MessageGui m = new MessageGui(
                        "Texture message",
                        "Processing [  end  ]",
                        MessageGui.OK_MESSAGE);
                //</editor-fold>

                for (int i = 0; i < seriesNames.length; i++) {
                    // update min and max values
                    seriesModels[i].rawTZYX.updateMinMaxValues();

                    // update color model
                    ColorModelParam colorModelParam = new ColorModelParam(ColorModelParam.RAINBOW_PVALUES0);
                    ColorController.updateColorModelParamOnAllFrames(seriesModels[i], colorModelParam);
                    ColorController.setDefaultWindowLevelFromHistogram(seriesModels[i]);
                    PicImageBoxOptionController
                            .getPicImageBoxOptionFromSerie(seriesModels[i]).picImageJPanelMenu.mapScaleColorMenu
                            .colorModelChanged();

                    // refresh GUI
                    FrameController.setActivatedSeriesOnAllFrames(seriesModels[i]);
                    FrameController.allFramesUpdateTransformChanged_0(seriesModels[i]);
                    FrameController.allFramesUpdateColorModelChanged_1(seriesModels[i]);
                    FrameController.allFramesUpdateSampleModelChanged_2(seriesModels[i]);
                    FrameController.allFramesUpdateWritableRasterChanged_3(seriesModels[i]);
                    FrameController.allFramesUpdateBufferedImageChanged_4(seriesModels[i]);
                }

                // update curves
                CurveController.run(
                        CurveEnum.ALL_ROI, null,
                        CurveEnum.ALL_SERIES, null
                );

                FrameController.setActivatedSeriesOnAllFrames(seriesModels[3]);//GLCM_Entropy_log10

            }
        } catch (Exception ex) {
            FrameGui.LOGGER.log(Level.INFO, "TexMainSub:: error {0} ", ex);
        }

        long end_time = System.currentTimeMillis();
        long difference = end_time - start_time;

        FrameGui.LOGGER.log(
                Level.INFO,
                "Time calculation of textural map proceding: {0} sec",
                (difference / 1_000d)
        );
        taskEnd();
        messageGui.end();
    }

    @Override
    public String toString() {
        String text;

        text = "Series / ROI parameters:"
                + "\nSeries processing:\t" + series.seriesDescription
                + "\nSeries modality:\t\t" + series.modality;

        text += "\n\nTextural parameters:"
                + "\nSpatial Resampling       -> Z spatial spacing:\t" + this.spatialResamplingZ
                + "\nSpatial Resampling       -> Y spatial spacing:\t" + this.spatialResamplingY
                + "\nSpatial Resampling       -> X spatial spacing:\t" + this.spatialResamplingX
                + "\nIntensity Rescaling      -> Min Bound:\t" + this.intensityRescalingMinBound
                + "\nIntensity Rescaling      -> Max Bound:\t" + this.intensityRescalingMaxBound
                + "\nIntensity Discretization -> Number of Grey Levels:\t" + intensityDiscretizationNbGreyLevels
                + "\nIntensity Discretization -> Bin Size:\t" + intensityDiscretizationBinSize;

        text += "\n\nTextural features:";
        text += "\n" + texGLCMMap.toString();

        return text;
    }

    /**
     * <p>
     * @return
     */
    private boolean isCheckTask() {

        //<editor-fold defaultstate="collapsed" desc="check intensityResampling min max values">
        if (FrameGui.getExecutionModel() != FrameGui.ExecutionModel.EXEC_SCRIPTING_TEXTURE) {
            if (texGLCMMapJPanelOption.jRadioButtonAbsoluteMinMax.isSelected()) {
                if (this.intensityRescalingMinBound == 0f
                        && this.intensityRescalingMaxBound == 0f) {
                    MessageGui messageGUI = new MessageGui(
                            "Texture Map",
                            "The bound (Min or Max) value is not defined.",
                            MessageGui.ERROR_MESSAGE);
                    return false;
                }
            }
        }
        //</editor-fold>

        //<editor-fold defaultstate="collapsed" desc="check intensityDiscretization NbGreyLevels value">
        if (FrameGui.getExecutionModel() != FrameGui.ExecutionModel.EXEC_SCRIPTING_TEXTURE) {
            if (this.intensityDiscretizationNbGreyLevels < 1) {
                MessageGui messageGUI = new MessageGui(
                        "Texture Protocol",
                        "The nbGreyLevels value is not defined or cannot be smaller than 1.",
                        MessageGui.ERROR_MESSAGE);
                return false;
            }
        }
        //</editor-fold>

        return true;
    }

    public TextureGLCMMap texGLCMMap;

    public double intensityDiscretizationNbGreyLevels;
    public double intensityDiscretizationBinSize;

    private double intensityRescalingMinBound;
    private double intensityRescalingMaxBound;

    public double rawZYXCropDiscretMaxValue;
    public double rawZYXCropDiscretMinValue;

    public float spatialResamplingZ;
    public float spatialResamplingY;
    public float spatialResamplingX;

    /**
     * get information about clustering for xls user row
     */
    public String informationClusterStr = "";

    private final SeriesModel series;

    private boolean isPT_NM = true;

    public final TextureLocalMapJPanel texGLCMMapJPanelOption;

    /**
     * kernel size
     */
    int kernelSize;

    protected static final int NB_DIRECTIONS_3D = 13;
    protected static final int[][] DIRECTIONS_3D = {
        //1  2   3  4  5   6  7  8   9  10  11  12  13
        {1, 1, 1, 1, 1, 1, 0, 0, 0, -1, -1, -1, 0}, //x
        {0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0}, //y
        {1, 0, -1, 1, 0, -1, 1, 0, -1, 1, 0, -1, 1} //z
    };

    protected static final int NB_DIRECTIONS_2D = 4;
    protected static final int[][] DIRECTIONS_2D = {
        //1  2  3   4 
        {1, 1, 0, -1}, //x
        {0, 1, 1, 1}, //y
        {0, 0, 0, 0} //z
    };

    /**
     * ON_ROI or ON_WHOLE_BODY model
     */
    private final TexGLCMMapModel model;

}
