/*
 * USE OF THE LIFEx SOFTWARE IS SUBJECT TO THE FOLLOWING TERMS AND CONDITIONS. 
 * IF YOU DO NOT ACCEPT THESE TERMS AND CONDITIONS, DO NOT DOWNLOAD THE SOFTWARE. 
 * NON-PROFESSIONAL USERS ARE NOTIFIED THAT THE SOFTWARE AND THE RISKS INHERENT TO ITS USE,
 * MODIFICATION AND/OR DEVELOPMENT AND REPRODUCTION ARE NOT COVERED BY ANY WARRANTY ON THE PART OF THE CEA. 
 * IT IS HIGHLY RECOMMENDED THAT YOU ONLY LOAD AND/OR USE THE SOFTWARE IF YOU HAVE SPECIALIZED IN-DEPTH INFORMATION TECHNOLOGY KNOW-HOW.
 * THE COMMISSARIAT A L’ENERGIE ATOMIQUE ET AUX ENERGIES ALTERNATIVES (FRENCH ATOMIC ENERGY AGENCY), 
 * a public scientific, technical and industrial establishment, 
 * having its registered office at 25 rue Leblanc – bâtiment le Ponant D – 75015 Paris cedex, 
 * hereinafter referred to as "CEA", hereby licenses the right to use the LIFEx software as per the terms and conditions set out below.  
 */
package bardp.protocol.protocolTexture.traditional;

import bardp.gui.FrameGui;
import bardp.protocol.protocolTexture.traditional.ConventionalPeak.ConvPeakParams;
import bardp.protocol.protocolTexture.params.TextureJPanelOutputParameters;
import bardp.raw.RawSparseZYX;
import bardp.raw.RawTZYX;
import bardp.roi.calc.RoiPeakModel;
import bardp.session.patient.study.series.SeriesModel;
import bardp.session.patient.study.series.SeriesUnits;
import bardp.session.patient.study.series.load.loadDcm.PT_NM.PT_NM;
import bardp.utility.Utils;
import java.util.logging.Level;
import org.apache.commons.math3.exception.MathIllegalArgumentException;
import org.apache.commons.math3.stat.descriptive.rank.Percentile;

/**
 * <p>
 * Title: Local Image Feature Extraction (LIFEx)</p>
 * <p>
 * Description: </p>
 * <p>
 * Copyright: Copyright (c) 2016</p>
 * <p>
 * Company: Inserm/CNRS/CEA</p>
 * <p>
 * @author Christophe Nioche, Inserm at lifexsoft.org
 * @version 6.0
 */
public class TraditionalController {

    public TraditionalController(
            final SeriesModel seriesRaw,
            final RawTZYX rawZYX,
            final RawSparseZYX roiZYXSparse
    ) {
        this.seriesRaw = seriesRaw;
        this.rawZYX = rawZYX;
        this.roiZYXSparse = roiZYXSparse;
    }

    public void run(
            double[] values
    ) {
        final int count = values.length;

        TextureJPanelOutputParameters texJPanelOutputParameters
                = FrameGui.protocolFrame.texBoxOption.textureJPanelOption.texGlobalROIJPanel.texJFrameOutputParameters.texJPanelOutputParameters;

        //<editor-fold defaultstate="collapsed" desc="Conventional_Q1_Q2_Q3">
        if (texJPanelOutputParameters.jCheckBoxConventional_Q1_Q2_Q3.isSelected()) {

            Percentile percentile = new Percentile();
            try {
                percentile25Value = percentile.evaluate(values, 25);
            } catch (MathIllegalArgumentException ignore) {
                percentile25Value = Double.NaN;
            }
            try {
                percentile50Value = percentile.evaluate(values, 50);
            } catch (MathIllegalArgumentException ignore) {
                percentile25Value = Double.NaN;
            }
            try {
                percentile75Value = percentile.evaluate(values, 75);
            } catch (MathIllegalArgumentException ignore) {
                percentile25Value = Double.NaN;
            }
        } else {
            percentile25Value = Double.NaN;
            percentile50Value = Double.NaN;
            percentile75Value = Double.NaN;
        }
        //</editor-fold>

        //<editor-fold defaultstate="collapsed" desc="mean std min max">
        if (texJPanelOutputParameters.jCheckBoxConventional_SUVmin_mean_max.isSelected()
                || texJPanelOutputParameters.jCheckBoxConventional_TLGmL.isSelected()) {

            long nbVox = 0;
            double m2 = 0f;
            double delta;

            minValue = Double.POSITIVE_INFINITY;
            maxValue = Double.NEGATIVE_INFINITY;

            for (double value : values) {
                nbVox++;
                delta = value - meanValue;
                meanValue += delta / nbVox;
                m2 += delta * (value - meanValue);

                if (value < minValue) {
                    minValue = value;
                }
                if (value > maxValue) {
                    maxValue = value;
                }
            }

            stdValue = Math.sqrt(m2 / nbVox);

            if (minValue == Double.POSITIVE_INFINITY) {
                minValue = 0d;
            }
            if (maxValue == Double.NEGATIVE_INFINITY) {
                maxValue = 0d;
            }
     
            volume_vx = nbVox;
            volume_mL = nbVox * rawZYX.iopIppSp.spModel.getVoxelSpacing() / 1_000d;
        } else {
            minValue = Double.NaN;
            meanValue = Double.NaN;
            stdValue = Double.NaN;
            maxValue = Double.NaN;
        }
        //</editor-fold>

        //<editor-fold defaultstate="collapsed" desc="volume + TLG">
        if (texJPanelOutputParameters.jCheckBoxConventional_TLGmL.isSelected()) {
            tlg = volume_mL * meanValue;
        }
        //</editor-fold>

        //<editor-fold defaultstate="collapsed" desc="picROISUVPeak1mL">
        if (texJPanelOutputParameters.jCheckBoxConventional_SUVpeakSphere1mL.isSelected()) {
            ConventionalPeak convPeak1mL = new ConventionalPeak(
                    rawZYX,
                    roiZYXSparse,
                    ConvPeakParams.PEAK_VOL1CM3
            );
            RoiPeakModel peakModel1mL = convPeak1mL.run();
            try {
                this.peak1mLDiscretizedVolumeSought = peakModel1mL.peakDiscretizedVolumeSought;
                this.peak1mLValue = peakModel1mL.peakValue;
                this.peak1mLStd = peakModel1mL.peakStdev;
                FrameGui.LOGGER.log(Level.INFO, "\n*_peakModel1mL: {0}+/-{1}", new Object[]{peak1mLValue, peak1mLStd});
            } catch (NullPointerException ignore) {
            }
        }
        //</editor-fold>

        //<editor-fold defaultstate="collapsed" desc="picROISUVPeak05mL">
        if (texJPanelOutputParameters.jCheckBoxConventional_SUVpeakSphere05mL.isSelected()) {
            ConventionalPeak convPeak05mL = new ConventionalPeak(
                    rawZYX,
                    roiZYXSparse,
                    ConvPeakParams.PEAK_VOL05CM3
            );
            RoiPeakModel peakModel05mL = convPeak05mL.run();
            try {
                this.peak05mLDiscretizedVolumeSought = peakModel05mL.peakDiscretizedVolumeSought;
                this.peak05mLValue = peakModel05mL.peakValue;
                this.peak05mLStd = peakModel05mL.peakStdev;
                FrameGui.LOGGER.log(Level.INFO, "\n*_peakModel05mL: {0}+/-{1}", new Object[]{peak05mLValue, peak05mLStd});
            } catch (NullPointerException ignore) {
            }
        }
        //</editor-fold>

        //<editor-fold defaultstate="collapsed" desc="percentile">
        //</editor-fold>
        if (texJPanelOutputParameters.jCheckBoxConventional_SUVmin_mean_max.isSelected()
                || texJPanelOutputParameters.jCheckBoxConventional_TLGmL.isSelected()) {

            double sumN, sumD;

            //<editor-fold defaultstate="collapsed" desc="skewness">
            // https://en.wikipedia.org/wiki/Skewness
            {
                sumN = 0d;
                sumD = 0d;
                for (double value : values) {
                    sumN = sumN + Math.pow(value - meanValue, 3);
                    sumD = sumD + Math.pow(value - meanValue, 2);
                }
                sumN /= (double) count;
                sumD /= (double) count;
                skewness = sumN / Math.pow(sumD, 1.5d);
                FrameGui.LOGGER.log(Level.INFO, "*_Skewness={0}", skewness);
            }
            //</editor-fold>

            //<editor-fold defaultstate="collapsed" desc="kurtosis">
            // https://en.wikipedia.org/wiki/Kurtosis
            {
                sumN = 0d;
                sumD = 0d;
                for (double value : values) {
                    sumN = sumN + Math.pow(value - meanValue, 4);
                    sumD = sumD + Math.pow(value - meanValue, 2);
                }
                sumN /= (double) count;
                sumD /= (double) count;
                kurtosis = sumN / Math.pow(sumD, 2);
                FrameGui.LOGGER.log(Level.INFO, "*_Kurtosis={0}", kurtosis);
            }
            //</editor-fold>

            //<editor-fold defaultstate="collapsed" desc="ExcessKurtosis">
            // HISTO_kurtosis is corrected by a Fisher correction of -3 to center kurtosis on 0 for normal distributions
            // IBSI protocol, p.33
            excessKurtosis = kurtosis - 3d;
            FrameGui.LOGGER.log(Level.INFO, "*_ExcessKurtosis={0}", excessKurtosis);
            //</editor-fold>
        }
    }

    @Override
    public String toString() {
        if (seriesRaw.seriesUnit.unitY == SeriesUnits.UNIT_SUV_BW) {
            return "\nSUVbw values:"
                    + "\n  SUVmean:\t\t" + Utils.d2s(getMeanValue(), 1)
                    + "\n  SUVstd:\t\t" + Utils.d2s(getStdValue(), 1)
                    + "\n  SUVmin:\t\t" + Utils.d2s(getMinValue(), 1)
                    + "\n  SUVmax:\t\t" + Utils.d2s(getMaxValue(), 1)
                    + "\n  SUVpeak (1cm3):\t" + Utils.d2s(getPeak1mLValue(), 1)
                    + "\n  SUVpeak (0.5cm3):\t" + Utils.d2s(getPeak05mLValue(), 1)
                    + "+-" + Utils.d2s(getPeak1mLStd(), 1)
                    + "\n"
                    + "\n  Volume (cm3):\t\t" + getVolume_mL()
                    + "\n  Volume (vx):\t\t" + getVolume_vx()
                    + "\n  TLG (mL):\t\t" + getTlg();
        } else if (seriesRaw.seriesUnit.unitY == SeriesUnits.UNIT_SUV_LBM) {
            return "\nSUVlbm values:"
                    + "\n  SULmean:\t\t" + Utils.d2s(getMeanValue(), 1)
                    + "\n  SULstd:\t\t" + Utils.d2s(getStdValue(), 1)
                    + "\n  SULmin:\t\t" + Utils.d2s(getMinValue(), 1)
                    + "\n  SULmax:\t\t" + Utils.d2s(getMaxValue(), 1)
                    + "\n  SULpeak (1cm3):\t" + Utils.d2s(getPeak1mLValue(), 1)
                    + "+-" + Utils.d2s(getPeak1mLStd(), 1)
                    + "\n  SULpeak (0.5cm3):\t" + Utils.d2s(getPeak05mLValue(), 1)
                    + "+-" + Utils.d2s(getPeak05mLStd(), 1)
                    + "\n"
                    + "\n  Volume (cm3):\t\t" + getVolume_mL()
                    + "\n  Volume (vx):\t\t" + getVolume_vx()
                    + "\n  TLG (mL):\t\t" + getTlg();
        } else if (seriesRaw.seriesUnit.unitY == SeriesUnits.UNIT_SUV_BSA) {
            return "\nSUVbsa values:"
                    + "\n  SULmean:\t\t" + Utils.d2s(getMeanValue(), 1)
                    + "\n  SULstd:\t\t" + Utils.d2s(getStdValue(), 1)
                    + "\n  SULmin:\t\t" + Utils.d2s(getMinValue(), 1)
                    + "\n  SULmax:\t\t" + Utils.d2s(getMaxValue(), 1)
                    + "\n  SULpeak (1cm3):\t" + Utils.d2s(getPeak1mLValue(), 1)
                    + "+-" + Utils.d2s(getPeak1mLStd(), 1)
                    + "\n  SULpeak (0.5cm3):\t" + Utils.d2s(getPeak05mLValue(), 1)
                    + "+-" + Utils.d2s(getPeak05mLStd(), 1)
                    + "\n"
                    + "\n  Volume (cm3):\t\t" + getVolume_mL()
                    + "\n  Volume (vx):\t\t" + getVolume_vx()
                    + "\n  TLG (mL):\t\t" + getTlg();
        } else if (seriesRaw.seriesUnit.unitY == SeriesUnits.UNIT_SUV_IBW) {
            return "\nSUVibw values:"
                    + "\n  SULmean:\t\t" + Utils.d2s(getMeanValue(), 1)
                    + "\n  SULstd:\t\t" + Utils.d2s(getStdValue(), 1)
                    + "\n  SULmin:\t\t" + Utils.d2s(getMinValue(), 1)
                    + "\n  SULmax:\t\t" + Utils.d2s(getMaxValue(), 1)
                    + "\n  SULpeak (1cm3):\t" + Utils.d2s(getPeak1mLValue(), 1)
                    + "+-" + Utils.d2s(getPeak1mLStd(), 1)
                    + "\n  SULpeak (0.5cm3):\t" + Utils.d2s(getPeak05mLValue(), 1)
                    + "+-" + Utils.d2s(getPeak05mLStd(), 1)
                    + "\n"
                    + "\n  Volume (cm3):\t\t" + getVolume_mL()
                    + "\n  Volume (vx):\t\t" + getVolume_vx()
                    + "\n  TLG (mL):\t\t" + getTlg();
        } else if (seriesRaw.seriesUnit.unitY == SeriesUnits.UNIT_kBqmL) {
            return "\nkBq/mL values:"
                    + "\n  kBq/mLmean:\t\t" + Utils.d2s(getMeanValue(), 1)
                    + "\n  kBq/mLstd:\t\t" + Utils.d2s(getStdValue(), 1)
                    + "\n  kBq/mLmin:\t\t" + Utils.d2s(getMinValue(), 1)
                    + "\n  kBq/mLmax:\t\t" + Utils.d2s(getMaxValue(), 1)
                    + "\n  kBq/mLpeak (1cm3):\t" + Utils.d2s(getPeak1mLValue(), 1)
                    + "+-" + Utils.d2s(getPeak1mLStd(), 1)
                    + "\n  kBq/mLpeak (0.5cm3):\t" + Utils.d2s(getPeak05mLValue(), 1)
                    + "+-" + Utils.d2s(getPeak05mLStd(), 1)
                    + "\n"
                    + "\n  Volume (cm3):\t\t" + getVolume_mL()
                    + "\n  Volume (vx):\t\t" + getVolume_vx()
                    + "\n  TLG (mL):\t\t" + getTlg();
        } else {
            return "\n Values:"
                    + "\n  mean:\t\t" + Utils.d2s(getMeanValue(), 2)
                    + "\n  std:\t\t" + Utils.d2s(getStdValue(), 2)
                    + "\n  min:\t\t" + Utils.d2s(getMinValue(), 2)
                    + "\n  max:\t\t" + Utils.d2s(getMaxValue(), 2)
                    + "\n  peak (1cm3):\t" + Utils.d2s(getPeak1mLValue(), 2)
                    + "+-" + Utils.d2s(getPeak1mLStd(), 2)
                    + "\n  peak (0.5cm3):\t" + Utils.d2s(getPeak05mLValue(), 2)
                    + "+-" + Utils.d2s(getPeak05mLStd(), 2)
                    + "\n"
                    + "\n  Volume (cm3):\t\t" + getVolume_mL()
                    + "\n  Volume (vx):\t\t" + getVolume_vx()
                    + "\n  TLG (mL):\t\t" + getTlg();
        }
    }

    public double getPeak1mLStd() {
        return peak1mLStd;
    }

    /**
     * Discretized Volume Sought of 1mL peak for peak (mm3)
     *
     * @return
     */
    public double getPeak1mLDiscretizedVolumeSought() {
        return peak1mLDiscretizedVolumeSought;
    }

    public double getPeak1mLValue() {
        return peak1mLValue;
    }

    public double getPeak05mLStd() {
        return peak05mLStd;
    }

    /**
     * Discretized Volume Sought of 0.5mL peak for peak (mm3)
     *
     * @return
     */
    public double getPeak05mLDiscretizedVolumeSought() {
        return peak05mLDiscretizedVolumeSought;
    }

    public double getPeak05mLValue() {
        return peak05mLValue;
    }

    public double getMaxValue() {
        return maxValue;
    }

    public double getMeanValue() {
        return meanValue;
    }

    public double getStdValue() {
        return stdValue;
    }

    public double getMinValue() {
        return minValue;
    }

    public double getVolume_mL() {
        return volume_mL;
    }

    public double getTlg() {
        return tlg;
    }

    public double getVolume_vx() {
        return volume_vx;
    }

    public double getPercentile25Value() {
        return percentile25Value;
    }

    public double getPercentile50Value() {
        return percentile50Value;
    }

    public double getPercentile75Value() {
        return percentile75Value;
    }

    public double getSkewness() {
        return skewness;
    }

    public double getKurtosis() {
        return kurtosis;
    }

    public double getExcessKurtosis() {
        return excessKurtosis;
    }

    /**
     * mean value
     */
    private double meanValue;

    /**
     * standard deviation value
     */
    private double stdValue;

    /**
     * minimum value
     */
    private double minValue;

    /**
     * maximum value
     */
    private double maxValue;

    /**
     * first quartile (Q1)
     */
    private double percentile25Value;

    /**
     * second quartile (Q2) is the median
     */
    private double percentile50Value;

    /**
     * third quartile (Q3)
     */
    private double percentile75Value;

    private double volume_mL;
    private double volume_vx;
    private double tlg;

    private double peak1mLValue;

    /**
     * Discretized Volume Sought of 1mL peak for peak (mm3)
     */
    private double peak1mLDiscretizedVolumeSought;
    private double peak1mLStd;

    private double peak05mLValue;

    /**
     * Discretized Volume Sought of 0.5mL peak for peak (mm3)
     */
    private double peak05mLDiscretizedVolumeSought;
    private double peak05mLStd;

    private double skewness;  //french: Asymétrie

    /**
     * kurtosis is calculated as measure of peakedness of the distribution
     */
    private double kurtosis;

    /**
     * excess Kurtosis is corrected by a Fisher correction of -3 to center kurtosis on 0 for normal distributions
     */
    private double excessKurtosis;

    private final RawTZYX rawZYX;
    private final RawSparseZYX roiZYXSparse;
    private final SeriesModel seriesRaw;
}
